/*
 * Hormann, Leydold, Derflinger
 * Automatic Nonuniform Random Variate Generation
 * Springer, Heidelberg, New York, 2003
 * http://statistik.wu-wien.ac.at/arvag/monograph
 * 
 * Code for Algorithm 11.4 (Cholesky-Decomopsition, p. 251)
 *
 * this file: algo11_04.c
 *
 * For a demo how to use this algorithm see the file "demo_algo11_03.c".
 *
 */

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include "uniform.h"


#define I(i,j) ((i)*(d+1)+(j))

int algo11_04(int d, double *S, double *L)
/*
 * Algorithm 11.4 (Cholesky-Decomopsition, p. 251)
 *
 * d ... dimension (length) of the generated vectors
 * S ... pointer to the array that holds the consecutive columns 
 *       of the variance-covariance matrix
 * L ... pointer to the array that holds (as result of the algorithm)
 *       the consecutive rows of the Cholesky-factor
 * 
 * Output: 1 is returned if it was possible to calculate the Cholesky factor.
 *         0 if the matrix was not positive definite.
 */
{      int i,j,k;
       double sum,x;

/*1*/  for(i=1;i<=d;i++)
/*2*/	 L[I(i,1)] = S[I(i,1)]/sqrt(S[I(1,1)]);
/*3*/  for(i=2;i<=d;i++){
/*4*/	 for(j=2;j<i;j++){
/*5*/	   sum=0.;
	   for(k=1;k<j;k++) sum+=L[I(i,k)]*L[I(j,k)];
	   L[I(i,j)] = (S[I(i,j)]-sum)/L[I(j,j)];
	 }
/*6*/	 sum = 0.;
	 for(j=1;j<i;j++) sum+=(L[I(i,j)])*(L[I(i,j)]);
	 x = S[I(i,i)]-sum;
/*7*/	 if(x>=0.) 
/*8*/	   L[I(i,i)] = sqrt(x);
/*9*/	 else{
/*10*/	   printf("ERROR Algo11_04: Sigma-matrix not positive definite!!\n");
	   printf("Problem occured for i=%d x%f\n",i,x);
	   return(0);
	 }
       }
/*11*/ for(i=1;i<=d;i++)
	 for(j=i+1;j<=d;j++) L[I(i,j)] = 0;
/*12*/ return 1;
} 
#undef I
